/*
 * File:   main.c
 * Author: Alberto Gmez Blzquez
 * Email: alberto.gomez.blazquez@estudiant.upc.edu
 *
 * Proyecto de estacin meteorolgica de bajo consumo e inalmbrica (Bluetooth)
 * Microcontrolador PIC18F87K22
 * Fosc = 4MHz
 * Envio de datos mediante protocolo RS232 (USART) y un dispositivo bluetooth HC-05 al PC (Interfaz en LabView)
 * Modo Sleep durante 1 minuto con despertar por desbordamiento de perro guardian (WDT)

 * Descripcin de los sensores y el protocolo de comunicacin que utilizan
 * ======================================================================================
 * Sensor------Protocolo-----------------Descricin--------------------------------------
 * ======================================================================================
 * DS18B20     1-Wire                    Sensor para medir la temperatura del agua
 * DHT22       Propio (Similar 1-Wire)   Sensor de humedad relativa y temperatura ambiente
 * BMP180      I2C                       Sensor para medir la presin atmoefrica y la temperatura ambiente
 * Anemmetro  Interrupcin por pulsos   Anemmetro que genera 1 pulsacin por cada vuelta
 * WindVane    Divisor de Tensin (ADC)  Genera una tensin diferente para cada posicin. Se mide con el ADC del micro
 * 
 */


#include <xc.h>
#include <stdio.h>
#include <string.h>
#include <math.h>
#include "delay.h"
#include "BMP180.h"
#include "DHT22.h"
#include "1wire.h"
#include "USART.h"
#include "Windvane.h"
#include "Windspeed.h"
#include <pic18f87k22.h>
#include <plib/i2c.h>
#include <plib/usart.h>
#include <plib/delays.h>

/* 	=================================================================================
        Definitions
    ================================================================================= */
// Definicin de los puertos
#define 	PORTBIT(adr, bit) ( (unsigned)(&adr)*8 + (bit) )

//Definicin de los estados
#define 	Idle            'A'
#define 	Setup           'B'
#define 	Read_Cal_BMP180 'C'
#define 	Read_DHT22      'D'
#define     Read_BMP180     'E'
#define     Read_DS18B20    'F'
#define     Read_Windvane   'G'
#define     Read_Vel_wind   'H'
#define     Send_data       'I'
#define     Sleep_mode      'J'
#define     Stop            'K'

//Definicin de los osciladores para configurar correctamente los delays

#define     LF                      0
#define     MF                      1
#define     HF                      2
#define     PLL                     3
#define     XT_10                   4
#define     _XTAL_FREQ              4000000
unsigned char Freq;

//Bus i2c
unsigned char sync_mode = 0, slew = 0;

//USART
char Txdata0[50] = "0";
unsigned char Rxdata;
unsigned char config=0,spbrg=0,baudconfig=0, i=0;
unsigned char inicio_usart = 1;


//Sensor BMP180
char TxdataBMP180[25] = "0";

//Sensor DHT22
char TxdataDHT22[15] = "0";
unsigned char DHT_22_flag;

//Sensor DS18B20
char TxdataDS18B20[16];
char TxdataDS18B20_USART[25] = "0";

//Anemmetro
char TxdataAnemometro[5] = "0";
unsigned int Pulsos;
unsigned char Time_out;
unsigned int NumPulsos;

//Windvane
char TxdataWindvane[3] ="0";

/* 	=================================================================================
        Global variables
    ================================================================================= */
//Leds de control placa
static bit LED_2 @ PORTBIT(PORTJ, 0);
static bit LED_1 @ PORTBIT(PORTD, 7);

//Variables estado inicial FSM
static char present_state = Idle;

//Variable interruocin RB0
unsigned char INT0_ST_SP_Flag;

//Variables interrupcin USART
unsigned char Go_Idle_Flag;
unsigned char Go_Flag;


/* 	=================================================================================
        Function prototypes
    ================================================================================= */

void init_system(void);
static void interrupt service_routine(void);
char state_logic(void);
char output_logic(void);
void Setup_I2C(void);
void Setup_USART_1(void);
void setup_timer_0(void);


/* 	=================================================================================
        Main function
    ================================================================================= */

void main(void) 
{
  //Programa principal
    init_system();
     /* loop forever */
    while (1) {
        output_logic();
        state_logic();
    }
}

/* 	=================================================================================
        Functions
    ================================================================================= */
void init_system(void) {
// Configuracin bits de control PIC18F87K22
    
// CONFIG1L
#pragma config RETEN = OFF      // VREG Sleep Enable bit (Disabled - Controlled by SRETEN bit)
#pragma config INTOSCSEL = LOW  // LF-INTOSC Low-power Enable bit (LF-INTOSC in Low-power mode during Sleep)
#pragma config SOSCSEL = HIGH   // SOSC Power Selection and mode Configuration bits (Low Power SOSC circuit selected)
#pragma config XINST = OFF      // Extended Instruction Set (Disabled)

// CONFIG1H
#pragma config FOSC = INTIO2    // Oscillator (Internal RC oscillator)
#pragma config PLLCFG = OFF     // PLL x4 Enable bit (Disabled)
#pragma config FCMEN = OFF      // Fail-Safe Clock Monitor (Disabled)
#pragma config IESO = OFF       // Internal External Oscillator Switch Over Mode (Disabled)

// CONFIG2L
#pragma config PWRTEN = OFF     // Power Up Timer (Disabled)
#pragma config BOREN = OFF      // Brown Out Detect (Disabled in hardware, SBOREN disabled)
#pragma config BORV = 3         // Brown-out Reset Voltage bits (1.8V)
#pragma config BORPWR = ZPBORMV // BORMV Power level (ZPBORMV instead of BORMV is selected)

// CONFIG2H
#pragma config WDTEN = ON      // Watchdog Timer (WDT disabled in hardware; SWDTEN bit disabled)
#pragma config WDTPS = 16384    // Watchdog Postscaler (1:16384) 65 seg

// CONFIG3L
    
#pragma config RTCOSC = SOSCREF // RTCC Clock Select (RTCC uses SOSC)
#pragma config EASHFT = ON      // External Address Shift bit (Address Shifting enabled)
#pragma config ABW = MM         // Address Bus Width Select bits (8-bit address bus)
#pragma config BW = 16          // Data Bus Width (16-bit external bus mode)
#pragma config WAIT = OFF       // External Bus Wait (Disabled)

// CONFIG3H
#pragma config CCP2MX = PORTC   // CCP2 Mux (RC1)
#pragma config ECCPMX = PORTH   // ECCP Mux (Enhanced CCP1/3 [P1B/P1C/P3B/P3C] muxed with RE6/RE5/RE4/RE3)
#pragma config MSSPMSK = MSK7   // MSSP address masking (7 Bit address masking mode)
#pragma config MCLRE = ON       // Master Clear Enable (MCLR Enabled, RG5 Disabled)

// CONFIG4L
#pragma config STVREN = ON      // Stack Overflow Reset (Enabled)
#pragma config BBSIZ = BB2K     // Boot Block Size (2K word Boot Block size)

// CONFIG5L
#pragma config CP0 = OFF        // Code Protect 00800-03FFF (Disabled)
#pragma config CP1 = OFF        // Code Protect 04000-07FFF (Disabled)
#pragma config CP2 = OFF        // Code Protect 08000-0BFFF (Disabled)
#pragma config CP3 = OFF        // Code Protect 0C000-0FFFF (Disabled)
#pragma config CP4 = OFF        // Code Protect 10000-13FFF (Disabled)
#pragma config CP5 = OFF        // Code Protect 14000-17FFF (Disabled)
#pragma config CP6 = OFF        // Code Protect 18000-1BFFF (Disabled)
#pragma config CP7 = OFF        // Code Protect 1C000-1FFFF (Disabled)

// CONFIG5H
#pragma config CPB = OFF        // Code Protect Boot (Disabled)
#pragma config CPD = OFF        // Data EE Read Protect (Disabled)

// CONFIG6L
#pragma config WRT0 = OFF       // Table Write Protect 00800-03FFF (Disabled)
#pragma config WRT1 = OFF       // Table Write Protect 04000-07FFF (Disabled)
#pragma config WRT2 = OFF       // Table Write Protect 08000-0BFFF (Disabled)
#pragma config WRT3 = OFF       // Table Write Protect 0C000-0FFFF (Disabled)
#pragma config WRT4 = OFF       // Table Write Protect 10000-13FFF (Disabled)
#pragma config WRT5 = OFF       // Table Write Protect 14000-17FFF (Disabled)
#pragma config WRT6 = OFF       // Table Write Protect 18000-1BFFF (Disabled)
#pragma config WRT7 = OFF       // Table Write Protect 1C000-1FFFF (Disabled)

// CONFIG6H
#pragma config WRTC = OFF       // Config. Write Protect (Disabled)
#pragma config WRTB = OFF       // Table Write Protect Boot (Disabled)
#pragma config WRTD = OFF       // Data EE Write Protect (Disabled)

// CONFIG7L
#pragma config EBRT0 = OFF      // Table Read Protect 00800-03FFF (Disabled)
#pragma config EBRT1 = OFF      // Table Read Protect 04000-07FFF (Disabled)
#pragma config EBRT2 = OFF      // Table Read Protect 08000-0BFFF (Disabled)
#pragma config EBRT3 = OFF      // Table Read Protect 0C000-0FFFF (Disabled)
#pragma config EBRT4 = OFF      // Table Read Protect 10000-13FFF (Disabled)
#pragma config EBRT5 = OFF      // Table Read Protect 14000-17FFF (Disabled)
#pragma config EBRT6 = OFF      // Table Read Protect 18000-1BFFF (Disabled)
#pragma config EBRT7 = OFF      // Table Read Protect 1C000-1FFFF (Disabled)

// CONFIG7H
#pragma config EBRTB = OFF      // Table Read Protect Boot (Disabled)
    

    
//Configuracin de las resistencias de pull-up en los puertos PORTB

    INTCON2bits.RBPU = 0; // 0 = Habilitadas las resistencias de pull-up en PORTB
                          // 1 = Deshabilitadas las resistencias de pull-up en PORTB

//Configuracin del oscilador interno
//|IDLEN|IRCF2|IRCF1|IRFC0|OSTS|HFIOFS|SCS1|SCS0|
//|  0  |  1  |  0  |  1  |  1 |   1  |  0 |  0 |
//HF (4MHz), PLL=OFF,(IDLEN=0)Modo sleep activado, Modo Idle desactivado
    OSCCON = 0b01011100;
    OSCCON2bits.MFIOFS = 0;
    OSCCON2bits.MFIOSEL = 0;
    OSCTUNEbits.PLLEN = 0;
    OSCTUNEbits.INTSRC = 0;

//Configuracin de los puertos 
    
// PORTA:
    PORTA = 0x00; // Reset all Flip-Flops at PORTA
    LATA = 0x00; // Reset all Flip-Flops at PORTA
    TRISA = 0b11011010; // RA5 LED D6 /RA2 Alimentacin del sensor de temperatura
    ANCON0bits.ANSEL2 = 0; //Puerto RA2 salida digital
   
    
// PORTB:
    PORTB = 0x00; // Reset all Flip-Flops at PORTB
    LATB = 0x00; // Reset all Flip-Flops at PORTB
    TRISB = 0b11001111; // RB0 BOTON/ RB2 Anemmetro/ RB4 LED D7/ RB5 LED D8
    
// PORTC:
    PORTC = 0x00; // Reset all Flip-Flops at PORTC
    LATC = 0x00; // Reset all Flip-Flops at PORTC
    TRISC = 0b10011000; //RC7 Rx(USART), RC6 Tx (USART), RC4=SDA1(I2C), RC3=SCL1(I2C)

// PORTD:
    PORTD = 0X00; // Reset all Flip-Flops at PORTD
    LATD = 0X00; // Reset all Flip-Flops at PORTD
    TRISD = 0b01100010; //RD6=SCL(I2C), RD5=SDA(I2C)
    
// PORTE:
    PORTE = 0x00; // Reset all Flip-Flops at PORTE
    LATE = 0x00; // Reset all Flip-Flops at PORTE
    TRISE = 0b00001111; // RE7/6/5/4 LED D4D3D2D1

// PORTF:
    PORTF = 0x00; // Reset all Flip-Flops at PORTE
    LATF = 0x00; // Reset all Flip-Flops at PORTE
    TRISF = 0b00000010; // RE7/6/5/4 LED D4D3D2D1
    ANCON0bits.ANSEL6 = 1; //Habilitada entrada analgica en puerto RF1
    ANCON0bits.ANSEL7 = 0; // I/O Digital en puerto RF2
    ANCON1bits.ANSEL8 = 0; // I/O Digital en puerto RF3
    
// PORTH:
    PORTH = 0x00; // Reset all Flip-Flops at PORTH
    LATH = 0x00; // Reset all Flip-Flops at PORTH
    TRISH = 0b00000000; // Todo salidas
    ANCON1 = 0b00000000; //Pines I/O se configuran como digitales
    ANCON2 = 0b00000000;
    
// PORTJ:
    PORTJ = 0x00; // Reset all Flip-Flops at PORTE
    LATJ = 0x00; // Reset all Flip-Flops at PORTE
    TRISJ = 0b00000000; // RE7/6/5/4 LED D4D3D2D1
    
// Configuracin inicial ADC
    ADCON1bits.VCFG1=0; // Vref+ = Vdd
    ADCON1bits.VCFG0=0;
    ADCON1bits.VNCFG=0; // Vref- = Vss
    ADCON2 = 0b10010100; //ADRES justificado derecha, 4TAD, Fosc/4
    
// Configuracin INTCON
    GIE = 1; // Global interrupts allowed
    PEIE = 1; //Interrupts periferic
    
// The INT2
    INTEDG2 = 0; // INT2 edge selection: '0' --> on falling edge (interrupt after clicking the pushbutton)
    INT2IE = 1; // Enable interrupts from the INT2 pin
    
//Configuracin de la frecuencia para los delays
    Freq = HF;
    
//Configuracin del reloj para i2c
    SSP1ADD = 0x09; 
            
//Condicin inicial de las variables
    
    INT0_ST_SP_Flag = 0;
    Go_Idle_Flag = 0;
    Go_Flag = 0;
    Time_out = 1;
    NumPulsos = 0;
    present_state = Idle;
    
}
//*****************************************************************************
//Future_state_logic routine
//*****************************************************************************

char state_logic(void) {

    char error = 0;
    
    switch (present_state) {
        case Idle:
            if (Go_Flag == 1) { 
                present_state = Setup;
                Go_Flag = 0;
            } 
            else
                present_state = Idle;
            break;

        case Setup:
             if (Go_Flag == 1) { 
                present_state = Stop;
                Go_Flag = 1;
            } else
                present_state = Read_Cal_BMP180;
            break;
            
        case Read_Cal_BMP180:
            if (Go_Flag == 1) { 
                present_state = Stop;
                Go_Flag = 1;
            } else
                present_state = Read_DHT22;
            break;

        case Read_DHT22:
            if (Go_Flag == 1 ) {
                present_state = Stop;
                Go_Flag = 0;
            } else
                present_state = Read_BMP180;
            break;

        case Read_BMP180:
            if (Go_Flag == 1 ) {
                present_state = Stop;
                Go_Flag = 0;
            } else
                present_state = Read_DS18B20;
            break;
        
        case Read_DS18B20:
            if (Go_Flag == 1) { 
                present_state = Stop;
                Go_Flag = 0;
            } else
                present_state = Read_Windvane;
            break;
            
        case Read_Windvane:
            if (Go_Flag == 1) { 
                present_state = Stop;
                Go_Flag = 0;
            } else
                present_state = Read_Vel_wind;
            break;
            

        case Read_Vel_wind:
            if (Go_Flag == 1) { 
                present_state = Stop;
                Go_Flag = 0;
            } else
                present_state = Send_data;
            break;

        case Send_data:
            if (Go_Flag == 1) { 
                present_state = Stop;
                Go_Flag = 0;
            } else
            present_state = Sleep_mode;
            break;
            
        case Sleep_mode:
            if (Go_Idle_Flag == 1) { 
                present_state = Stop;
                Go_Idle_Flag = 0;
            } else
            present_state = Read_DHT22;
            break;
            
        case Stop:
            present_state = Idle;
            break;
            
        default:
            error = 1;
    }
    return (error);
}

//*****************************************************************************
//Output logic routine
//*****************************************************************************

char output_logic(void) {
    char error = 0;
    
    switch (present_state) {
        case Idle:
            //Inicio modulo USART
            if (inicio_usart == 1) {
                Setup_USART_1();
            }    
            //USART se inicia solo una vez
            inicio_usart = 0;            
            //Leds control placa que inidican que estamos en modo Idle
            LED_1 = 1; LED_2 = 1; 
         
            break;
            
        case Setup:   
            //Setup I2C
            //100kHz Baud clock(SSPADD =9) Fosc=4MHz Configurado en Init_System
            Setup_I2C();           
            //Preparo el bus dels sensor DHT22
            Setup_port_DHT22();          
            //Preparo el bus dels sensor DS18B20
            Setup_1_wire_port();
                     
            break;
            
        case Read_Cal_BMP180:
            //Cambio estado leds control para indicar que estamos obteniendo y enviando datos
            LED_1 = 0; LED_2 = 1; 
            //Leer parametros de calibracin del sensor BMP180
            Get_CalValue_BMP180();
            
            break;

        case Read_DHT22:          
            //Medimos la humedad y la temperatura ambiente con el sensor DHT22
            Read_Dev_DHT22();
            
            break;
            
        case Read_BMP180:
            //Medimos la temperatura y la presin atmosfrica con el sensor I2C BMP180
            Read_Dev_BMP180();    
                       
            break;
            
        case Read_DS18B20:     
            //Leo el sensor impermeable de temperatura DS18B20
            ReadTemp_DS18B20(); 

            break;
        
        case Read_Windvane:         
            //Capturo direccin del viento con ADC
            Capture_Windvane();
            
            break;
            
        case Read_Vel_wind:
            //Cuento los pulsos de anemmetro durante 2.4 seg para medir la velocidad del viento
            Read_VelWind_Int();
                     
            break;
        
        case Send_data:
            //Juntar los datos y enviarlos por USART           
            strcpy(Txdata0,TxdataDHT22);
            strcat(Txdata0,TxdataBMP180);
            strcat(Txdata0,TxdataDS18B20);
            strcat(Txdata0,TxdataAnemometro);
            strcat(Txdata0,TxdataWindvane);
            strcat(Txdata0,"\r\n");        
            //Enviar datos  por puerto serie      
            while (Busy1USART()); //Check if Usart is busy or not
            puts1USART((char *) Txdata0); //transmit the string
            
            break;
            
        case Sleep_mode:
            //Entro en modo sleep
            //Leds que indican modo Sleep
            LED_1 = 1; LED_2 = 0; 
       
            //WDT Setup
            //Watchdog timer is ON       
            WDTCONbits.SWDTEN = 1;
            BAUDCON1bits.WUE = 1; //Configuracin la interrupcin EUSART en modo sleep
            //Entro en modo Sleep     
            Sleep();
            NOP();//Se recomienda que la siguiente instruccin despus del modo sleep sea NOP
            
            //El dato recibido en modo Sleep es indiferente, solo activa la bandera de dato recibido
            Rxdata = getc1USART(); //Recibo el dato despus del modo sleep 
            
            __delay_ms(20);
            // Si el estado de el registro WUE ha cambiado a 0, ha habido una interrupcin por USART
            if (BAUDCON1bits.WUE == 0) {
                Go_Idle_Flag = 1;
            }
            //Limpio la bandera de recepcin USART
            PIR1bits.RC1IF = 0;
            //Watchdog timer is OFF
            WDTCONbits.SWDTEN = 0;   
            
            break;
            
        case Stop:
            //Cierro comunicacin USART
            Close1USART();
            inicio_usart = 1; //Para volver a enviar mensaje de inicio
            //Apago puerto DHT22
            Off_port_DHT22();
            //Apago puerto DS18B20
            Off_1_wire_port();
            
            break;            
            
        default:
            error = 1;
    }
    return (error);
}

/* 	=================================================================================
        Interrupt function
    ================================================================================= */
static void interrupt service_routine(void) { //Interrupt is the key word here
GIE = 0; //Disable interrupts while attending one of them ...

//Interrupcin del Timer0 cada 2.4seg
if (TMR0IF == 1) {
    TMR0 = 46786; //Precargo de nuevo el valor 46786
    Pulsos = NumPulsos;
    NumPulsos = 0;
    Time_out = 0; //Flag de tiempo transcurrido de 2.4 segundos
    TMR0IF = 0;
}

//Interrupcin de anammetro en RB2
if (INT2IF == 1) {
    NumPulsos++;
    INT2IF = 0;
}

//Interrupcin USART
if (RC1IF == 1) {
    Rxdata = getc1USART(); //Se lee el dato que esta en el buffer de Rx del USART
    PIR1bits.RC1IF = 0; //Desactivamos la bandera de recepcin en el buffer de entrada del USART
    
    
    switch (Rxdata) {
//Datos recibidos en estado IDLE
       
            case '3':
                Go_Flag = 1;
                break;
            default:
                Go_Flag = 0;
                break;
        }
    Rxdata = 0; 
}

GIE = 1; //Enable interrupts
}

//Funciones I2C
void Setup_I2C(void) {
    sync_mode = MASTER;
    slew = SLEW_OFF;
    OpenI2C(sync_mode, slew);
    
}

//Timer 0
void setup_timer_0(void) {
//    Configuracin Timer0 para tiempo 2.4 segundos con Fosc=4MHz:
//            |Tiempo|Prescaler|T0PS2|T0PS1|T0PS0|  TMR0  |
//            |2.4seg|   128   |  1  |  1  |  0  | 46786  |
//    
//    IMPORTANTE!!:Hay que modificar el valor de TMR0 en la interrupcin del timer0 
//    Para cualquier otro tiempo:
//      Tiempo = (4/Fosc)*prescaler*(65536-TMR0)  
    TMR0IE = 1; // Enable TMR0 interrupts
    PSA = 0; // Prescaler in use
    T0CONbits.T08BIT = 0; //Mode 16-bits
    T0CONbits.T0CS = 0; // clock = FOSC/4
    T0PS0 = 0;
    T0PS1 = 1;
    T0PS2 = 1; // T0PS(2..0) = "110" programs a 1:128 Prescale value
    TMR0 = 46786; //Preload 46786 at TMR0 count TMR0=(65536-46786)=18570  
}



