/*
 * File:   main.c
 * Author: Alberto Gmez Blzquez
 * alberto.gomez.blazquez@estudiant.upc.edu
 *
 * Proyecto para medir la humedad y la temperatura con el sensor DHT22 (Protocolo similar al 1-wire)
 * Microcontrolador PIC18F87K22
 * Fosc = 4MHz
 * Descripcin de los sensores y el protocolo de comunicacin que utilizan
 * ======================================================================================
 * Sensor------Protocolo-----------------Descricin--------------------------------------
 * ======================================================================================
 * DHT22       Propio (Similar 1-Wire)   Sensor de humedad relativa y temperatura 
 * 
 */
 


#include <xc.h>
#include <stdio.h>
#include <pic18f87k22.h>
#include <plib/delays.h>
#include <plib/xlcd.h>
#include <plib/usart.h>
#include "delay.h"
#include "DHT22.h"
#include "USART.h"
/* 	=================================================================================
        Definitions
    ================================================================================= */
// Definicin de los puertos
#define 	PORTBIT(adr, bit) ( (unsigned)(&adr)*8 + (bit) )

//Definicin de los estados
#define 	Idle            'A'
#define 	Setup           'B'
#define 	Read_DHT22      'C'
#define 	Stop            'D'
#define     Error_DHT22     'E'

//Definicin de los osciladores para configurar correctamente los delays

#define     LF                      0
#define     MF                      1
#define     HF                      2
#define     PLL                     3
#define     XT_10                   4
#define     _XTAL_FREQ              4000000
unsigned char Freq;


/* 	=================================================================================
        Global variables
    ================================================================================= */
static bit ST_SP @ PORTBIT(PORTB, 0);
static bit LED_7 @ PORTBIT(PORTB, 5);
static bit LED_6 @ PORTBIT(PORTB, 4);
static bit LED_5 @ PORTBIT(PORTA, 5);
static bit LED_4 @ PORTBIT(PORTE, 7);
static bit LED_3 @ PORTBIT(PORTE, 6);
static bit LED_2 @ PORTBIT(PORTE, 5);
static bit LED_1 @ PORTBIT(PORTE, 4);



//Variables estado inicial FSM
static char present_state = Idle;

//Variable interruocin RB0
unsigned char INT0_ST_SP_Flag;

//Sensor DHT22
char TxdataDHT22[15];
char TxdataDHT22_USART[50];
unsigned char DHT_22_flag;
char mostrar_Temp[16];
char mostrar_Hum [16];

//Variables para el LCD
unsigned char LCD_Flag;

//Variables para USART
unsigned char inicio_usart = 1;


/* 	=================================================================================
        Function prototypes
    ================================================================================= */

void init_system(void);
static void interrupt service_routine(void);
char state_logic(void);
char output_logic(void);
void DelayFor18TCY(void);
void DelayPORXLCD(void);
void DelayXLCD(void);
void init_XLCD(void);
void welcome_lcd(void);
void Hum_Temp_XLCD(void);
void error_dht22_lcd(void);



/* 	=================================================================================
        Main function
    ================================================================================= */

void main(void) 
{
  //Programa principal
    init_system();
     /* loop forever */
    while (1) {
        output_logic();
        state_logic();
    }
}

/* 	=================================================================================
        Functions
    ================================================================================= */
void init_system(void) {
// Configuracin bits de control PIC18F87K22
    
// CONFIG1L
#pragma config RETEN = OFF      // VREG Sleep Enable bit (Disabled - Controlled by SRETEN bit)
#pragma config INTOSCSEL = LOW  // LF-INTOSC Low-power Enable bit (LF-INTOSC in Low-power mode during Sleep)
#pragma config SOSCSEL = HIGH    // SOSC Power Selection and mode Configuration bits (Low Power SOSC circuit selected)
#pragma config XINST = OFF      // Extended Instruction Set (Disabled)

// CONFIG1H
//#pragma config FOSC = XT        // Oscillator (XT oscillator)
#pragma config FOSC = INTIO2    // Oscillator (Internal RC oscillator)
#pragma config PLLCFG = OFF     // PLL x4 Enable bit (Disabled)
#pragma config FCMEN = OFF      // Fail-Safe Clock Monitor (Disabled)
#pragma config IESO = OFF       // Internal External Oscillator Switch Over Mode (Disabled)

// CONFIG2L
#pragma config PWRTEN = ON      // Power Up Timer (Enabled)
#pragma config BOREN = SBORDIS  // Brown Out Detect (Enabled in hardware, SBOREN disabled)
#pragma config BORV = 3         // Brown-out Reset Voltage bits (1.8V)
#pragma config BORPWR = ZPBORMV // BORMV Power level (ZPBORMV instead of BORMV is selected)

// CONFIG2H
#pragma config WDTEN = OFF      // Watchdog Timer (WDT disabled in hardware; SWDTEN bit disabled)
#pragma config WDTPS = 1048576  // Watchdog Postscaler (1:1048576)

// CONFIG3L
#pragma config RTCOSC = SOSCREF // RTCC Clock Select (RTCC uses SOSC)
#pragma config EASHFT = ON      // External Address Shift bit (Address Shifting enabled)
#pragma config ABW = MM         // Address Bus Width Select bits (8-bit address bus)
#pragma config BW = 16          // Data Bus Width (16-bit external bus mode)
#pragma config WAIT = OFF       // External Bus Wait (Disabled)

// CONFIG3H
#pragma config CCP2MX = PORTC   // CCP2 Mux (RC1)
#pragma config ECCPMX = PORTH   // ECCP Mux (Enhanced CCP1/3 [P1B/P1C/P3B/P3C] muxed with RE6/RE5/RE4/RE3)
#pragma config MSSPMSK = MSK7   // MSSP address masking (7 Bit address masking mode)
#pragma config MCLRE = ON       // Master Clear Enable (MCLR Enabled, RG5 Disabled)

// CONFIG4L
#pragma config STVREN = ON      // Stack Overflow Reset (Enabled)
#pragma config BBSIZ = BB2K     // Boot Block Size (2K word Boot Block size)

// CONFIG5L
#pragma config CP0 = OFF        // Code Protect 00800-03FFF (Disabled)
#pragma config CP1 = OFF        // Code Protect 04000-07FFF (Disabled)
#pragma config CP2 = OFF        // Code Protect 08000-0BFFF (Disabled)
#pragma config CP3 = OFF        // Code Protect 0C000-0FFFF (Disabled)
#pragma config CP4 = OFF        // Code Protect 10000-13FFF (Disabled)
#pragma config CP5 = OFF        // Code Protect 14000-17FFF (Disabled)
#pragma config CP6 = OFF        // Code Protect 18000-1BFFF (Disabled)
#pragma config CP7 = OFF        // Code Protect 1C000-1FFFF (Disabled)

// CONFIG5H
#pragma config CPB = OFF        // Code Protect Boot (Disabled)
#pragma config CPD = OFF        // Data EE Read Protect (Disabled)

// CONFIG6L
#pragma config WRT0 = OFF       // Table Write Protect 00800-03FFF (Disabled)
#pragma config WRT1 = OFF       // Table Write Protect 04000-07FFF (Disabled)
#pragma config WRT2 = OFF       // Table Write Protect 08000-0BFFF (Disabled)
#pragma config WRT3 = OFF       // Table Write Protect 0C000-0FFFF (Disabled)
#pragma config WRT4 = OFF       // Table Write Protect 10000-13FFF (Disabled)
#pragma config WRT5 = OFF       // Table Write Protect 14000-17FFF (Disabled)
#pragma config WRT6 = OFF       // Table Write Protect 18000-1BFFF (Disabled)
#pragma config WRT7 = OFF       // Table Write Protect 1C000-1FFFF (Disabled)

// CONFIG6H
#pragma config WRTC = OFF       // Config. Write Protect (Disabled)
#pragma config WRTB = OFF       // Table Write Protect Boot (Disabled)
#pragma config WRTD = OFF       // Data EE Write Protect (Disabled)

// CONFIG7L
#pragma config EBRT0 = OFF      // Table Read Protect 00800-03FFF (Disabled)
#pragma config EBRT1 = OFF      // Table Read Protect 04000-07FFF (Disabled)
#pragma config EBRT2 = OFF      // Table Read Protect 08000-0BFFF (Disabled)
#pragma config EBRT3 = OFF      // Table Read Protect 0C000-0FFFF (Disabled)
#pragma config EBRT4 = OFF      // Table Read Protect 10000-13FFF (Disabled)
#pragma config EBRT5 = OFF      // Table Read Protect 14000-17FFF (Disabled)
#pragma config EBRT6 = OFF      // Table Read Protect 18000-1BFFF (Disabled)
#pragma config EBRT7 = OFF      // Table Read Protect 1C000-1FFFF (Disabled)

// CONFIG7H
#pragma config EBRTB = OFF      // Table Read Protect Boot (Disabled)
    

//Configuracin de las resistencias de pull-up en los puertos PORTB

    INTCON2bits.RBPU = 0; // 0 = Habilitadas las resistencias de pull-up en PORTB
                          // 1 = Deshabilitadas las resistencias de pull-up en PORTB

//Configuracin del oscilador interno
//|IDLEN|IRCF2|IRCF1|IRFC0|OSTS|HFIOFS|SCS1|SCS0|
//|  0  |  1  |  0  |  1  |  1 |   1  |  0 |  0 |
//HF (4MHz), PLL=OFF, 
    OSCCON = 0b01011100;
    OSCCON2bits.MFIOFS = 0;
    OSCCON2bits.MFIOSEL = 0;
    OSCTUNEbits.PLLEN = 0;
    OSCTUNEbits.INTSRC = 0;

//Configuracin de los puertos 
    
// PORTA:
    PORTA = 0x00; // Reset all Flip-Flops at PORTA
    LATA = 0x00; // Reset all Flip-Flops at PORTA
    TRISA = 0b11011011; // RA5 LED D6 
    
// PORTB:
    PORTB = 0x00; // Reset all Flip-Flops at PORTB
    LATB = 0x00; // Reset all Flip-Flops at PORTB
    TRISB = 0b11001111; // RB0 BOTON/ RB4 LED D7/ RB5 LED D8
    
// PORTC:
    PORTC = 0x00; // Reset all Flip-Flops at PORTC
    LATC = 0x00; // Reset all Flip-Flops at PORTC
    TRISC = 0b10011000; //RC7 Rx(USART), RC6 Tx (USART)
    
// PORTE:
    PORTE = 0x00; // Reset all Flip-Flops at PORTE
    LATE = 0x00; // Reset all Flip-Flops at PORTE
    TRISE = 0b00001111; // RE7/6/5/4 LED D4D3D2D1

// PORTF:
    PORTF = 0x00; // Reset all Flip-Flops at PORTE
    LATF = 0x00; // Reset all Flip-Flops at PORTE
    TRISF = 0b00000010; // RE7/6/5/4 LED D4D3D2D1
    ANCON1bits.ANSEL8 = 0; // I/O Digital en puerto RF3
    
// PORTH:
    PORTH = 0x00; // Reset all Flip-Flops at PORTH
    LATH = 0x00; // Reset all Flip-Flops at PORTH
    TRISH = 0b00000000; // Todo salidas
    ANCON1 = 0b00000000; //Pines I/O se configuran como digitales
    ANCON2 = 0b00000000;

// The INT0
    INTEDG0 = 0; // INT0 edge selection: '0' --> on falling edge (interrupt after clicking the pushbutton)
    INT0IE = 1; // Enable interrupts from the INT0 pin
    
// Configuracin INTCON
    GIE = 1; // Global interrupts allowed
    PEIE = 1; //Interrupts periferic

//Configuracin de la frecuencia para los delays
    Freq = HF;
      
//Condicin inicial de las variables
    
    INT0_ST_SP_Flag = 0;
    LED_7 = 0;
    LED_6 = 0;
    LED_5 = 0;
    LED_4 = 0;
    LED_3 = 0;
    LED_2 = 0;
    LED_1 = 0;
    LCD_Flag = 1;
    present_state = Idle;
 
    //Iniciamos el LCD
    init_XLCD();
    
}
//*****************************************************************************
//Future_state_logic routine
//*****************************************************************************

char state_logic(void) {

    char error = 0;
    
    switch (present_state) {
        case Idle:
            if (INT0_ST_SP_Flag == 1) { // if a button pressed is detected,
                present_state = Setup;
                INT0_ST_SP_Flag = 0; //Clear variable, so that it ca be interrupted again
            } else
                present_state = Idle;
            break;

        case Setup:
                present_state = Read_DHT22;
            break;

        case Read_DHT22:

            if (INT0_ST_SP_Flag == 1) { // if a button pressed is detected,
                present_state = Stop;
                INT0_ST_SP_Flag = 0; //Clear variable, so that it ca be interrupted again
            } else if (DHT_22_flag == 1) {
                present_state = Error_DHT22;
                DHT_22_flag = 0; //Borramos la variable de error del sensor DHT22
            } else
                present_state = Read_DHT22;
            break;
            
        case Stop:
                present_state = Idle;
            break;
            
        case Error_DHT22:
            present_state = Read_DHT22;
            break;
            
        default:
            error = 1;
    }
    return (error);
}

//*****************************************************************************
//Output logic routine
//*****************************************************************************

char output_logic(void) { 
    char error = 0;
    
    switch (present_state) {
        case Idle:           
            if (LCD_Flag == 1){
            welcome_lcd();
            }
            //Flag para el cambio de mensaje
            LCD_Flag = 0;
            LED_7 = 1; LED_6 = 0; LED_5 = 1; LED_4 = 0; LED_3 = 1; LED_2 = 0; LED_1 = 1;
            Delay_mSec(1000);
            LED_7 = 0; LED_6 = 1; LED_5 = 0; LED_4 = 1; LED_3 = 0; LED_2 = 1; LED_1 = 0;
            Delay_mSec(1000);
            break;
            
        case Setup:                 
            //Borramos pantalla LCD
            WriteCmdXLCD(0x01);       
            //Preparo el bus dels sensor DHT22
            Setup_port_DHT22();
            //Inicio modulo USART
            if (inicio_usart == 1) {
                Setup_USART_1();
            }    
            //USART se inicia solo una vez
            inicio_usart = 0; 
            break;

        case Read_DHT22:
            LED_7 = 0; LED_6 = 1; LED_5 = 0; LED_4 = 0; LED_3 = 0; LED_2 = 0; LED_1 = 0;
            //Medimos la humedad y la temperatura ambiente con el sensor DHT22
            Read_Dev_DHT22();
//            Envio la informacin a la pantalla LCD
            if (LCD_Flag == 0) {
                Hum_Temp_XLCD();
            }
            //Enviar datos por USART      
            while (Busy1USART()); //Check if Usart is busy or not
            puts1USART((char *)TxdataDHT22_USART); //transmit the string
            //La consulta del sensor de humedad y temperatura DHT22 se tiene que realizar cada 2 segundos 0,5Hz
            Delay_Sec(1);
            Delay_Sec(1);
          
            break;
            
        case Stop:           
            LED_7 = 0; LED_6 = 0; LED_5 = 1; LED_4 = 1; LED_3 = 0; LED_2 = 1; LED_1 = 0;
            //Borramos pantalla LCD
            WriteCmdXLCD(0x01);
            //Flag para cambio de mensaje en LCD
            LCD_Flag = 1;
            //Apagamos el mdulo de comunicacin del sensor DHT22
            Off_port_DHT22();
            //Cierro comunicacin USART
            Close1USART();
            //Flag para volver a iniciar USART
            inicio_usart = 1; 
            break;
            
        case Error_DHT22:
            //Si el sensor no est disponible mostramos por pantalla durante un segundo el error
            if (LCD_Flag == 1) {
                error_dht22_lcd();
                Delay_Sec(1);
                //Borramos pantalla LCD
                WriteCmdXLCD(0x01);         
            }
            break;

        default:
            error = 1;
    }
    return (error);
}

/* 	=================================================================================
        Interrupt function
    ================================================================================= */
static void interrupt service_routine(void) { //Interrupt is the key word here
GIE = 0; //Disable interrupts while attending one of them ...
//Interrupcin Bontn RB1
if (INT0IF == 1) {
    INT0_ST_SP_Flag = 1; //Set the state diagram variable and clear the interrupt flag
    // (so that the system can be interrupted again)
    INT0IF = 0;
}

GIE = 1; //Enable interrupts
}

//Definicion de las funciones Delays para utilizar la libreria XLCD para una frecuencia de FOSC de 4MHz

void init_XLCD(void)                //Initialize LCD display
{
    OpenXLCD(FOUR_BIT&LINES_5X7);  //configure LCD in 4-bit Data Interface mode
                                     //and 5x7 characters, multiple line display
    while(BusyXLCD());             //Check if the LCD controller is not busy
                                     //before writing some commands?
    WriteCmdXLCD(0x06);            // move cursor right, don?t shift display
    WriteCmdXLCD(0x0C);            //turn display on without cursor

}
void Hum_Temp_XLCD(void){
    while (BusyXLCD()); //Check if the LCD controller is not busy
    SetDDRamAddr(0x00); //shift cursor to beginning of first line
    putrsXLCD("HUM(HR)  TEMP(C)"); //Display "Humedad" y "Temperatura"
    SetDDRamAddr(0x40); //shift cursor to beginning of second line
    putrsXLCD(TxdataDHT22); //Mostrar datos DHT22
}

void welcome_lcd(void) {
    while (BusyXLCD()); //Check if the LCD controller is not busy
    SetDDRamAddr(0x00); //shift cursor to beginning of first line
    putrsXLCD("Modo IDLE"); //Display "Modo IDLE"
    SetDDRamAddr(0x40); //shift cursor to beginning of second line
    putrsXLCD("Pulsa RB0"); //Display "Pulsa RB0"
}
void error_dht22_lcd(void) {
    while (BusyXLCD()); //Check if the LCD controller is not busy
    SetDDRamAddr(0x00); //shift cursor to beginning of first line
    putrsXLCD("Device DHT22"); //Display "Device DHT22"
    SetDDRamAddr(0x40); //shift cursor to beginning of second line
    putrsXLCD("Not Present"); //Display "Not Present"
}

void DelayFor18TCY( void )         //18 cycles delay
{
Delay10TCYx(20);
}

void DelayPORXLCD (void)          // Delay of 15ms
{
Delay1KTCYx(15);
}

void DelayXLCD (void)            // Delay of 5ms
{
Delay1KTCYx(5);
}

