/*
 * File:   main.c
 * Author: Alberto Gmez Blzquez
 * alberto.gomez.blazquez@estudiant.upc.edu
 *
 * Proyecto de medicin de presin atmosfrica y temperatura ambiente con el sensor BMP180
 * Envio de datos mediante protocolo RS232 (USART) 
 * Microcontrolador PIC18F87K22
 * Fosc = 4MHz
 * Descripcin de los sensores y el protocolo de comunicacin que utilizan
 * ======================================================================================
 * Sensor------Protocolo-----------------Descricin--------------------------------------
 * ======================================================================================
 * BMP180      I2C                       Sensor para medir la presin atmoefrica y la temperatura ambiente
 * 
 */


#include <xc.h>
#include <stdio.h>
#include <string.h>
#include <math.h>
#include "delay.h"
#include "BMP180.h"
#include "USART.h"
#include <pic18f87k22.h>
#include <plib/i2c.h>
#include <plib/usart.h>
#include <plib/delays.h>
#include <plib/xlcd.h>

/* 	=================================================================================
        Definitions
    ================================================================================= */
// Definicin de los puertos
#define 	PORTBIT(adr, bit) ( (unsigned)(&adr)*8 + (bit) )

//Definicin de los estados
#define 	Idle            'A'
#define 	Setup           'B'
#define 	Read_Cal_BMP180 'C'
#define     Read_BMP180     'D'
#define     Send_data       'E'
#define     Stop            'F'

//Definicin de los osciladores para configurar correctamente los delays

#define     LF                      0
#define     MF                      1
#define     HF                      2
#define     PLL                     3
#define     XT_10                   4
#define     _XTAL_FREQ              4000000
unsigned char Freq;

//Bus i2c
unsigned char sync_mode = 0, slew = 0;

//USART
char Txdata0[50] = "0";
unsigned char inicio_usart = 1;

//Sensor BMP180
char TxdataBMP180[25];
char TxdataBMP180_LCD[16];

//Variables para el LCD
unsigned char LCD_Flag;

/* 	=================================================================================
        Global variables
    ================================================================================= */
//Leds de control placa
static bit LED_7 @ PORTBIT(PORTB, 5);
static bit LED_6 @ PORTBIT(PORTB, 4);
static bit LED_5 @ PORTBIT(PORTA, 5);
static bit LED_4 @ PORTBIT(PORTE, 7);
static bit LED_3 @ PORTBIT(PORTE, 6);
static bit LED_2 @ PORTBIT(PORTE, 5);
static bit LED_1 @ PORTBIT(PORTE, 4);

//Variables estado inicial FSM
static char present_state = Idle;

//Variable interruocin RB0
unsigned char INT0_ST_SP_Flag;

/* 	=================================================================================
        Function prototypes
    ================================================================================= */

void init_system(void);
static void interrupt service_routine(void);
char state_logic(void);
char output_logic(void);
void Setup_I2C(void);
void Setup_USART_1(void);
void DelayFor18TCY(void);
void DelayPORXLCD(void);
void DelayXLCD(void);
void init_XLCD(void);
void welcome_lcd(void);
void Temp_XLCD(void);

/* 	=================================================================================
        Main function
    ================================================================================= */

void main(void) 
{
  //Programa principal
    init_system();
    present_state = Idle; //Descomentar para instalacin con PicKit3
     /* loop forever */
    while (1) {
        output_logic();
        state_logic();
    }
}

/* 	=================================================================================
        Functions
    ================================================================================= */
void init_system(void) {
// Configuracin bits de control PIC18F87K22
    
// CONFIG1L
#pragma config RETEN = OFF      // VREG Sleep Enable bit (Disabled - Controlled by SRETEN bit)
#pragma config INTOSCSEL = LOW  // LF-INTOSC Low-power Enable bit (LF-INTOSC in Low-power mode during Sleep)
#pragma config SOSCSEL = HIGH   // SOSC Power Selection and mode Configuration bits (Low Power SOSC circuit selected)
#pragma config XINST = OFF      // Extended Instruction Set (Disabled)

// CONFIG1H
//#pragma config FOSC = XT      // Oscillator (XT oscillator)
#pragma config FOSC = INTIO2    // Oscillator (Internal RC oscillator)
#pragma config PLLCFG = OFF     // PLL x4 Enable bit (Disabled)
#pragma config FCMEN = OFF      // Fail-Safe Clock Monitor (Disabled)
#pragma config IESO = OFF       // Internal External Oscillator Switch Over Mode (Disabled)

// CONFIG2L
#pragma config PWRTEN = ON      // Power Up Timer (Enabled)
#pragma config BOREN = SBORDIS  // Brown Out Detect (Enabled in hardware, SBOREN disabled)
#pragma config BORV = 3         // Brown-out Reset Voltage bits (1.8V)
#pragma config BORPWR = ZPBORMV // BORMV Power level (ZPBORMV instead of BORMV is selected)

// CONFIG2H
#pragma config WDTEN = OFF      // Watchdog Timer (WDT disabled in hardware; SWDTEN bit disabled)
#pragma config WDTPS = 16384    // Watchdog Postscaler (1:16384) 65 seg

// CONFIG3L
    
#pragma config RTCOSC = SOSCREF // RTCC Clock Select (RTCC uses SOSC)
#pragma config EASHFT = ON      // External Address Shift bit (Address Shifting enabled)
#pragma config ABW = MM         // Address Bus Width Select bits (8-bit address bus)
#pragma config BW = 16          // Data Bus Width (16-bit external bus mode)
#pragma config WAIT = OFF       // External Bus Wait (Disabled)

// CONFIG3H
#pragma config CCP2MX = PORTC   // CCP2 Mux (RC1)
#pragma config ECCPMX = PORTH   // ECCP Mux (Enhanced CCP1/3 [P1B/P1C/P3B/P3C] muxed with RE6/RE5/RE4/RE3)
#pragma config MSSPMSK = MSK7   // MSSP address masking (7 Bit address masking mode)
#pragma config MCLRE = ON       // Master Clear Enable (MCLR Enabled, RG5 Disabled)

// CONFIG4L
#pragma config STVREN = ON      // Stack Overflow Reset (Enabled)
#pragma config BBSIZ = BB2K     // Boot Block Size (2K word Boot Block size)

// CONFIG5L
#pragma config CP0 = OFF        // Code Protect 00800-03FFF (Disabled)
#pragma config CP1 = OFF        // Code Protect 04000-07FFF (Disabled)
#pragma config CP2 = OFF        // Code Protect 08000-0BFFF (Disabled)
#pragma config CP3 = OFF        // Code Protect 0C000-0FFFF (Disabled)
#pragma config CP4 = OFF        // Code Protect 10000-13FFF (Disabled)
#pragma config CP5 = OFF        // Code Protect 14000-17FFF (Disabled)
#pragma config CP6 = OFF        // Code Protect 18000-1BFFF (Disabled)
#pragma config CP7 = OFF        // Code Protect 1C000-1FFFF (Disabled)

// CONFIG5H
#pragma config CPB = OFF        // Code Protect Boot (Disabled)
#pragma config CPD = OFF        // Data EE Read Protect (Disabled)

// CONFIG6L
#pragma config WRT0 = OFF       // Table Write Protect 00800-03FFF (Disabled)
#pragma config WRT1 = OFF       // Table Write Protect 04000-07FFF (Disabled)
#pragma config WRT2 = OFF       // Table Write Protect 08000-0BFFF (Disabled)
#pragma config WRT3 = OFF       // Table Write Protect 0C000-0FFFF (Disabled)
#pragma config WRT4 = OFF       // Table Write Protect 10000-13FFF (Disabled)
#pragma config WRT5 = OFF       // Table Write Protect 14000-17FFF (Disabled)
#pragma config WRT6 = OFF       // Table Write Protect 18000-1BFFF (Disabled)
#pragma config WRT7 = OFF       // Table Write Protect 1C000-1FFFF (Disabled)

// CONFIG6H
#pragma config WRTC = OFF       // Config. Write Protect (Disabled)
#pragma config WRTB = OFF       // Table Write Protect Boot (Disabled)
#pragma config WRTD = OFF       // Data EE Write Protect (Disabled)

// CONFIG7L
#pragma config EBRT0 = OFF      // Table Read Protect 00800-03FFF (Disabled)
#pragma config EBRT1 = OFF      // Table Read Protect 04000-07FFF (Disabled)
#pragma config EBRT2 = OFF      // Table Read Protect 08000-0BFFF (Disabled)
#pragma config EBRT3 = OFF      // Table Read Protect 0C000-0FFFF (Disabled)
#pragma config EBRT4 = OFF      // Table Read Protect 10000-13FFF (Disabled)
#pragma config EBRT5 = OFF      // Table Read Protect 14000-17FFF (Disabled)
#pragma config EBRT6 = OFF      // Table Read Protect 18000-1BFFF (Disabled)
#pragma config EBRT7 = OFF      // Table Read Protect 1C000-1FFFF (Disabled)

// CONFIG7H
#pragma config EBRTB = OFF      // Table Read Protect Boot (Disabled)
    

    
//Configuracin de las resistencias de pull-up en los puertos PORTB

    INTCON2bits.RBPU = 0; // 0 = Habilitadas las resistencias de pull-up en PORTB
                          // 1 = Deshabilitadas las resistencias de pull-up en PORTB

//Configuracin del oscilador interno
//|IDLEN|IRCF2|IRCF1|IRFC0|OSTS|HFIOFS|SCS1|SCS0|
//|  0  |  1  |  0  |  1  |  1 |   1  |  0 |  0 |
//HF (4MHz), PLL=OFF,(IDLEN=0)Modo sleep activado, Modo Idle desactivado
    OSCCON = 0b01011100;
    OSCCON2bits.MFIOFS = 0;
    OSCCON2bits.MFIOSEL = 0;
    OSCTUNEbits.PLLEN = 0;
    OSCTUNEbits.INTSRC = 0;

//Configuracin de los puertos 
    
// PORTA:
    PORTA = 0x00; // Reset all Flip-Flops at PORTA
    LATA = 0x00; // Reset all Flip-Flops at PORTA
    TRISA = 0b11011010; // RA5 LED D6 
    ANCON0bits.ANSEL2 = 0; //Puerto RA2 salida digital
   
    
// PORTB:
    PORTB = 0x00; // Reset all Flip-Flops at PORTB
    LATB = 0x00; // Reset all Flip-Flops at PORTB
    TRISB = 0b11001111; // RB0 BOTON/ RB4 LED D7/ RB5 LED D8
    
// PORTC:
    PORTC = 0x00; // Reset all Flip-Flops at PORTC
    LATC = 0x00; // Reset all Flip-Flops at PORTC
    TRISC = 0b10011000; //RC7 Rx(USART), RC6 Tx (USART), RC4=SDA1(I2C), RC3=SCL1(I2C)

// PORTD:
    PORTD = 0X00; // Reset all Flip-Flops at PORTD
    LATD = 0X00; // Reset all Flip-Flops at PORTD
    TRISD = 0b01100010; //RD6=SCL(I2C), RD5=SDA(I2C)
    
// PORTE:
    PORTE = 0x00; // Reset all Flip-Flops at PORTE
    LATE = 0x00; // Reset all Flip-Flops at PORTE
    TRISE = 0b00001111; // RE7/6/5/4 LED D4D3D2D1

// PORTF:
    PORTF = 0x00; // Reset all Flip-Flops at PORTE
    LATF = 0x00; // Reset all Flip-Flops at PORTE
    TRISF = 0b00000010; // RE7/6/5/4 LED D4D3D2D1
    
// PORTH:
    PORTH = 0x00; // Reset all Flip-Flops at PORTH
    LATH = 0x00; // Reset all Flip-Flops at PORTH
    TRISH = 0b00000000; // Todo salidas
    ANCON1 = 0b00000000; //Pines I/O se configuran como digitales
    ANCON2 = 0b00000000;
    
// PORTJ:
    PORTJ = 0x00; // Reset all Flip-Flops at PORTE
    LATJ = 0x00; // Reset all Flip-Flops at PORTE
    TRISJ = 0b00000000; // RE7/6/5/4 LED D4D3D2D1
    
// Configuracin inicial ADC
    ADCON1bits.VCFG1=0; // Vref+ = Vdd
    ADCON1bits.VCFG0=0;
    ADCON1bits.VNCFG=0; // Vref- = Vss
    ADCON2 = 0b10010100; //ADRES justificado derecha, 4TAD, Fosc/4
    
// The INT0
    INTEDG0 = 0; // INT0 edge selection: '0' --> on falling edge (interrupt after clicking the pushbutton)
    INT0IE = 1; // Enable interrupts from the INT0 pin  
    
// Configuracin INTCON
    GIE = 1; // Global interrupts allowed
    PEIE = 1; //Interrupts periferic

//Configuracin de la frecuencia para los delays
    Freq = HF;
    
//Configuracin del reloj para i2c
    SSP1ADD = 0x09; 
    
//Condicin inicial de las variables
    
    INT0_ST_SP_Flag = 0;
    present_state = Idle;

//Iniciamos el LCD
    init_XLCD();
    LCD_Flag = 1;
    
}
//*****************************************************************************
//Future_state_logic routine
//*****************************************************************************

char state_logic(void) {

    char error = 0;
    
    switch (present_state) {
        case Idle:
            if (INT0_ST_SP_Flag == 1) { // if a button pressed is detected,
                present_state = Setup;
                INT0_ST_SP_Flag = 0;
            } 
            else
                present_state = Idle;
            break;

        case Setup:
          
                present_state = Read_Cal_BMP180;
            break;
            
        case Read_Cal_BMP180:
  
                present_state = Read_BMP180;

            break;

        case Read_BMP180:
            if (INT0_ST_SP_Flag == 1 ) { // if a button pressed is detected,
                present_state = Stop;
                INT0_ST_SP_Flag = 0;
            } else
                present_state = Send_data;
            break;
      
        case Send_data:
            if (INT0_ST_SP_Flag == 1) { // if a button pressed is detected,
                present_state = Stop;
                INT0_ST_SP_Flag = 0;
            } else
            present_state = Read_BMP180;
            break;
            
        case Stop:
            present_state = Idle;
            break;
            
        default:
            error = 1;
    }
    return (error);
}

//*****************************************************************************
//Output logic routine
//*****************************************************************************

char output_logic(void) {
    char error = 0;
    
    switch (present_state) {
        case Idle:
            if (LCD_Flag == 1){
            welcome_lcd();
            }
            LCD_Flag = 0;           
            //Leds control placa que inidican que estamos en modo Idle
            LED_7 = 1; LED_6 = 0; LED_5 = 1; LED_4 = 0; LED_3 = 1; LED_2 = 0; LED_1 = 1;
            Delay_mSec(1000);
            LED_7 = 0; LED_6 = 1; LED_5 = 0; LED_4 = 1; LED_3 = 0; LED_2 = 1; LED_1 = 0;
            Delay_mSec(1000);
         
            break;
            
        case Setup:   
            //Setup I2C
            //100kHz Baud clock(SSPADD =9) Fosc=4MHz Configurado en Init_System
            Setup_I2C();
            //Inicio modulo USART
            if (inicio_usart == 1) {
                Setup_USART_1();
            }
            //USART se inicia solo una vez
            inicio_usart = 0; 
            //Borramos pantalla LCD
            WriteCmdXLCD(0x01);
            break;
            
        case Read_Cal_BMP180:
            //Cambio estado leds control para indicar que estamos obteniendo y enviando datos
            LED_7 = 0; LED_6 = 1; LED_5 = 0; LED_4 = 1; LED_3 = 0; LED_2 = 1; LED_1 = 0;
            //Read Calibration parameters BMP180
            Get_CalValue_BMP180();
            
            break;
            
        case Read_BMP180:
            //Medimos la temperatura y la presin atmosfrica con el sensor I2C BMP180
            Read_Dev_BMP180();
                           
            break;

        case Send_data:
            //Juntar los datos y enviarlos por USART           
            strcpy(Txdata0,TxdataBMP180);
            strcat(Txdata0,"\r\n");        
            //Enviar datos  por puerto serie      
            while (Busy1USART()); //Check if Usart is busy or not
            puts1USART((char *) Txdata0); //transmit the string
            
            if (LCD_Flag == 0){
                Temp_XLCD();                         
            }
            Delay_mSec(1000);
            break;           
        
        case Stop:
            //Cierro comunicacin USART
            Close1USART();
            inicio_usart = 1; //Para volver a enviar mensaje de inicio
            LCD_Flag = 1;
            WriteCmdXLCD(0x01);
            break;            
            
        default:
            error = 1;
    }
    return (error);
}

/* 	=================================================================================
        Interrupt function
    ================================================================================= */
static void interrupt service_routine(void) { //Interrupt is the key word here
GIE = 0; //Disable interrupts while attending one of them ...

//Interrupcin Botn RB0
if (INT0IF == 1) {
    INT0_ST_SP_Flag = 1; //Set the state diagram variable and clear the interrupt flag
    // (so that the system can be interrupted again)
    INT0IF = 0;
}


GIE = 1; //Enable interrupts
}

//Funciones I2C
void Setup_I2C(void) {
    sync_mode = MASTER;
    slew = SLEW_OFF;
    OpenI2C(sync_mode, slew);
    
}

//Inicio LCD
void init_XLCD(void)                //Initialize LCD display
{
    OpenXLCD(FOUR_BIT&LINES_5X7);  //configure LCD in 4-bit Data Interface mode
                                     //and 5x7 characters, multiple line display
    while(BusyXLCD());             //Check if the LCD controller is not busy
                                     //before writing some commands?
    WriteCmdXLCD(0x06);            // move cursor right, don?t shift display
    WriteCmdXLCD(0x0C);            //turn display on without cursor

}

//Mensaje temperatura y presin atmosfrica LCD
void Temp_XLCD(void){
    while (BusyXLCD()); //Check if the LCD controller is not busy
    SetDDRamAddr(0x00); //shift cursor to beginning of first line
    putrsXLCD("PresAtm:  Temp:"); //Display "temperatura"
    SetDDRamAddr(0x40); //shift cursor to beginning of second line
    putrsXLCD(TxdataBMP180_LCD); //Display "Presin atm"
}

//Mensaje inicio LCD
void welcome_lcd(void) {
    while (BusyXLCD()); //Check if the LCD controller is not busy
    SetDDRamAddr(0x00); //shift cursor to beginning of first line
    putrsXLCD("Modo IDLE"); //Display "Modo IDLE"
    SetDDRamAddr(0x40); //shift cursor to beginning of second line
    putrsXLCD("Pulsa RB0"); //Display "Pulsa RB0"
}

//Funciones Delay para LCD
void DelayFor18TCY( void )         //18 cycles delay
{
Delay10TCYx(20);
}

void DelayPORXLCD (void)          // Delay of 15ms
{
Delay1KTCYx(15);
}

void DelayXLCD (void)            // Delay of 5ms
{
Delay1KTCYx(5);
}



